@@ -1,8 +1,7 @@
 from __future__ import print_function
+from __future__ import unicode_literals
 
 import os
-import subprocess
-import sys
 import textwrap
 import time
 from fnmatch import fnmatch
@@ -14,7 +13,7 @@ import six
 import tox
 from .config import parseconfig
 from .result import ResultLog
-from .session import Action
+from .session import main
 from .venv import VirtualEnv
 
 
@@ -32,7 +31,7 @@ def pytest_addoption(parser):
 
 
 def pytest_report_header():
-    return "tox comes from: %r" % (tox.__file__)
+    return "tox comes from: {}".format(repr(tox.__file__))
 
 
 @pytest.fixture
@@ -49,14 +48,59 @@ def newconfig(request, tmpdir):
             return parseconfig(args, plugins=plugins)
         finally:
             old.chdir()
+
     return newconfig
 
 
 @pytest.fixture
-def cmd(request):
+def cmd(request, capfd, monkeypatch):
     if request.config.option.no_network:
         pytest.skip("--no-network was specified, test cannot run")
-    return Cmd(request)
+    request.addfinalizer(py.path.local().chdir)
+
+    def run(*argv):
+        key = str(b'PYTHONPATH')
+        python_paths = (i for i in (str(os.getcwd()), os.getenv(key)) if i)
+        monkeypatch.setenv(key, os.pathsep.join(python_paths))
+        with RunResult(capfd, argv) as result:
+            try:
+                main([str(x) for x in argv])
+                assert False  # this should always exist with SystemExit
+            except SystemExit as exception:
+                result.ret = exception.code
+            except OSError as e:
+                result.ret = e.errno
+        return result
+
+    yield run
+
+
+class RunResult:
+    def __init__(self, capfd, args):
+        self._capfd = capfd
+        self.args = args
+        self.ret = None
+        self.duration = None
+        self.out = None
+        self.err = None
+
+    def __enter__(self):
+        self._start = time.time()
+        # noinspection PyProtectedMember
+        self._capfd._start()
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.duration = time.time() - self._start
+        self.out, self.err = self._capfd.readouterr()
+
+    @property
+    def outlines(self):
+        return self.out.splitlines()
+
+    def __repr__(self):
+        return 'RunResult(ret={}, args={}, out=\n{}\n, err=\n{})'.format(
+            self.ret, ' '.join(str(i) for i in self.args), self.out, self.err)
 
 
 class ReportExpectMock:
@@ -77,13 +121,9 @@ class ReportExpectMock:
 
         def generic_report(*args, **kwargs):
             self._calls.append((name,) + args)
-            print("%s" % (self._calls[-1], ))
-        return generic_report
+            print("%s" % (self._calls[-1],))
 
-    def action(self, venv, msg, *args):
-        self._calls.append(("action", venv, msg))
-        print("%s" % (self._calls[-1], ))
-        return Action(self.session, venv, msg, args)
+        return generic_report
 
     def getnext(self, cat):
         __tracebackhide__ = True
@@ -169,6 +209,7 @@ def mocksession(request):
             pm = pcallMock(args, cwd, env, stdout, stderr, shell)
             self._pcalls.append(pm)
             return pm
+
     return MockSession()
 
 
@@ -180,133 +221,68 @@ def newmocksession(request):
     def newmocksession(args, source, plugins=()):
         mocksession.config = newconfig(args, source, plugins=plugins)
         return mocksession
+
     return newmocksession
 
 
-class Cmd:
-    def __init__(self, request):
-        self.tmpdir = request.getfixturevalue("tmpdir")
-        self.request = request
-        current = py.path.local()
-        self.request.addfinalizer(current.chdir)
-
-    def chdir(self, target):
-        target.chdir()
-
-    def popen(self, argv, stdout, stderr, **kw):
-        env = os.environ.copy()
-        env['PYTHONPATH'] = ":".join(filter(None, [
-            str(os.getcwd()), env.get('PYTHONPATH', '')]))
-        kw['env'] = env
-        # print "env", env
-        return subprocess.Popen(argv, stdout=stdout, stderr=stderr, **kw)
-
-    def run(self, *argv):
-        if argv[0] == "tox" and sys.version_info[:2] < (2, 7):
-            pytest.skip("can not run tests involving calling tox on python2.6. "
-                        "(and python2.6 is about to be deprecated anyway)")
-        argv = [str(x) for x in argv]
-        assert py.path.local.sysfind(str(argv[0])), argv[0]
-        p1 = self.tmpdir.join("stdout")
-        p2 = self.tmpdir.join("stderr")
-        print("%s$ %s" % (os.getcwd(), " ".join(argv)))
-        f1 = p1.open("wb")
-        f2 = p2.open("wb")
-        now = time.time()
-        popen = self.popen(argv, stdout=f1, stderr=f2,
-                           close_fds=(sys.platform != "win32"))
-        ret = popen.wait()
-        f1.close()
-        f2.close()
-        out = p1.read("rb")
-        out = getdecoded(out).splitlines()
-        err = p2.read("rb")
-        err = getdecoded(err).splitlines()
-
-        def dump_lines(lines, fp):
-            try:
-                for line in lines:
-                    print(line, file=fp)
-            except UnicodeEncodeError:
-                print("couldn't print to %s because of encoding" % (fp,))
-        dump_lines(out, sys.stdout)
-        dump_lines(err, sys.stderr)
-        return RunResult(ret, out, err, time.time() - now)
-
-
 def getdecoded(out):
-        try:
-            return out.decode("utf-8")
-        except UnicodeDecodeError:
-            return "INTERNAL not-utf8-decodeable, truncated string:\n%s" % (
-                py.io.saferepr(out),)
-
-
-class RunResult:
-    def __init__(self, ret, outlines, errlines, duration):
-        self.ret = ret
-        self.outlines = outlines
-        self.errlines = errlines
-        self.stdout = LineMatcher(outlines)
-        self.stderr = LineMatcher(errlines)
-        self.duration = duration
-
-
-class LineMatcher:
-    def __init__(self, lines):
-        self.lines = lines
-
-    def str(self):
-        return "\n".join(self.lines)
-
-    def fnmatch_lines(self, lines2):
-        if isinstance(lines2, str):
-            lines2 = py.code.Source(lines2)
-        if isinstance(lines2, py.code.Source):
-            lines2 = lines2.strip().lines
-
-        from fnmatch import fnmatch
-        lines1 = self.lines[:]
-        nextline = None
-        extralines = []
-        __tracebackhide__ = True
-        for line in lines2:
-            nomatchprinted = False
-            while lines1:
-                nextline = lines1.pop(0)
-                if line == nextline:
-                    print("exact match:", repr(line))
-                    break
-                elif fnmatch(nextline, line):
-                    print("fnmatch:", repr(line))
-                    print("   with:", repr(nextline))
-                    break
-                else:
-                    if not nomatchprinted:
-                        print("nomatch:", repr(line))
-                        nomatchprinted = True
-                    print("    and:", repr(nextline))
-                extralines.append(nextline)
-            else:
-                assert line == nextline
+    try:
+        return out.decode("utf-8")
+    except UnicodeDecodeError:
+        return "INTERNAL not-utf8-decodeable, truncated string:\n%s" % (
+            py.io.saferepr(out),)
 
 
 @pytest.fixture
 def initproj(request, tmpdir):
-    """ create a factory function for creating example projects. """
+    """Create a factory function for creating example projects
+
+    Constructed folder/file hierarchy examples:
+
+    with `src_root` other than `.`:
+
+      tmpdir/
+          name/                  # base
+            src_root/            # src_root
+                name/            # package_dir
+                    __init__.py
+                name.egg-info/   # created later on package build
+            setup.py
+
+    with `src_root` given as `.`:
+
+      tmpdir/
+          name/                  # base, src_root
+            name/                # package_dir
+                __init__.py
+            name.egg-info/       # created later on package build
+            setup.py
+
+    """
+
     def initproj(nameversion, filedefs=None, src_root="."):
         if filedefs is None:
             filedefs = {}
+        if not src_root:
+            src_root = '.'
         if isinstance(nameversion, six.string_types):
-            parts = nameversion.split("-")
+            parts = nameversion.split(str("-"))
             if len(parts) == 1:
                 parts.append("0.1")
             name, version = parts
         else:
             name, version = nameversion
-        base = tmpdir.ensure(name, dir=1)
+
+        base = tmpdir.join(name)
+        src_root_path = _path_join(base, src_root)
+        assert base == src_root_path or src_root_path.relto(base), (
+            '`src_root` must be the constructed project folder or its direct '
+            'or indirect subfolder')
+
+        base.ensure(dir=1)
         create_files(base, filedefs)
-        if 'setup.py' not in filedefs:
+
+        if not _filedefs_contains(base, filedefs, 'setup.py'):
             create_files(base, {'setup.py': '''
                 from setuptools import setup, find_packages
                 setup(
@@ -319,24 +295,72 @@ def initproj(request, tmpdir):
                     package_dir={'':'%(src_root)s'},
                 )
             ''' % locals()})
-        if name not in filedefs:
-            src_dir = base.ensure(src_root, dir=1)
-            create_files(src_dir, {
+
+        if not _filedefs_contains(base, filedefs, src_root_path.join(name)):
+            create_files(src_root_path, {
                 name: {'__init__.py': '__version__ = %r' % version}
             })
-        manifestlines = []
-        for p in base.visit(lambda x: x.check(file=1)):
-            manifestlines.append("include %s" % p.relto(base))
+
+        manifestlines = ["include %s" % p.relto(base)
+                         for p in base.visit(lambda x: x.check(file=1))]
         create_files(base, {"MANIFEST.in": "\n".join(manifestlines)})
+
         print("created project in %s" % (base,))
         base.chdir()
+        return base
+
     return initproj
 
 
+def _path_parts(path):
+    path = path and str(path)  # py.path.local support
+    parts = []
+    while path:
+        folder, name = os.path.split(path)
+        if folder == path:  # root folder
+            folder, name = name, folder
+        if name:
+            parts.append(name)
+        path = folder
+    parts.reverse()
+    return parts
+
+
+def _path_join(base, *args):
+    # workaround for a py.path.local bug on Windows (`path.join('/x', abs=1)`
+    # should be py.path.local('X:\\x') where `X` is the current drive, when in
+    # fact it comes out as py.path.local('\\x'))
+    return py.path.local(base.join(*args, abs=1))
+
+
+def _filedefs_contains(base, filedefs, path):
+    """
+    whether `filedefs` defines a file/folder with the given `path`
+
+    `path`, if relative, will be interpreted relative to the `base` folder, and
+    whether relative or not, must refer to either the `base` folder or one of
+    its direct or indirect children. The base folder itself is considered
+    created if the filedefs structure is not empty.
+
+    """
+    unknown = object()
+    base = py.path.local(base)
+    path = _path_join(base, path)
+
+    path_rel_parts = _path_parts(path.relto(base))
+    for part in path_rel_parts:
+        if not isinstance(filedefs, dict):
+            return False
+        filedefs = filedefs.get(part, unknown)
+        if filedefs is unknown:
+            return False
+    return path_rel_parts or path == base and filedefs
+
+
 def create_files(base, filedefs):
     for key, value in filedefs.items():
         if isinstance(value, dict):
             create_files(base.ensure(key, dir=1), value)
-        elif isinstance(value, str):
+        elif isinstance(value, six.string_types):
             s = textwrap.dedent(value)
             base.join(key).write(s)
