@@ -34,10 +34,18 @@ def prepare(args):
     return config
 
 
-def main(args=None):
+def run_main(args=None):
+    if args is None:
+        args = sys.argv[1:]
+    main(args)
+
+
+def main(args):
     try:
         config = prepare(args)
         retcode = Session(config).runcommand()
+        if retcode is None:
+            retcode = 0
         raise SystemExit(retcode)
     except KeyboardInterrupt:
         raise SystemExit(2)
@@ -98,7 +106,10 @@ class Action(object):
 
     def setactivity(self, name, msg):
         self.activity = name
-        self.report.verbosity0("%s %s: %s" % (self.venvname, name, msg), bold=True)
+        if msg:
+            self.report.verbosity0("%s %s: %s" % (self.venvname, name, msg), bold=True)
+        else:
+            self.report.verbosity1("%s %s: %s" % (self.venvname, name, msg), bold=True)
 
     def info(self, name, msg):
         self.report.verbosity1("%s %s: %s" % (self.venvname, name, msg), bold=True)
@@ -109,12 +120,11 @@ class Action(object):
         else:
             logdir = self.session.config.logdir
         try:
-            l = logdir.listdir("%s-*" % actionid)
+            log_count = len(logdir.listdir("%s-*" % actionid))
         except (py.error.ENOENT, py.error.ENOTDIR):
             logdir.ensure(dir=1)
-            l = []
-        num = len(l)
-        path = logdir.join("%s-%s.log" % (actionid, num))
+            log_count = 0
+        path = logdir.join("%s-%s.log" % (actionid, log_count))
         f = path.open('w')
         f.flush()
         return f
@@ -126,8 +136,8 @@ class Action(object):
             fout = self._initlogpath(self.id)
             fout.write("actionid: %s\nmsg: %s\ncmdargs: %r\n\n" % (self.id, self.msg, args))
             fout.flush()
-            self.popen_outpath = outpath = py.path.local(fout.name)
-            fin = outpath.open()
+            outpath = py.path.local(fout.name)
+            fin = outpath.open('rb')
             fin.read()  # read the header, so it won't be written to stdout
             stdout = fout
         elif returnout:
@@ -151,21 +161,25 @@ class Action(object):
             self.report.logpopen(popen, env=env)
             try:
                 if resultjson and not redirect:
-                    assert popen.stderr is None  # prevent deadlock
+                    if popen.stderr is not None:
+                        # prevent deadlock
+                        raise ValueError("stderr must not be piped here")
+                    # we read binary from the process and must write using a
+                    # binary stream
+                    buf = getattr(sys.stdout, 'buffer', sys.stdout)
                     out = None
                     last_time = time.time()
                     while 1:
-                        fin_pos = fin.tell()
                         # we have to read one byte at a time, otherwise there
                         # might be no output for a long time with slow tests
                         data = fin.read(1)
                         if data:
-                            sys.stdout.write(data)
-                            if '\n' in data or (time.time() - last_time) > 1:
+                            buf.write(data)
+                            if b'\n' in data or (time.time() - last_time) > 1:
                                 # we flush on newlines or after 1 second to
                                 # provide quick enough feedback to the user
                                 # when printing a dot per test
-                                sys.stdout.flush()
+                                buf.flush()
                                 last_time = time.time()
                         elif popen.poll() is not None:
                             if popen.stdout is not None:
@@ -173,14 +187,15 @@ class Action(object):
                             break
                         else:
                             time.sleep(0.1)
-                            fin.seek(fin_pos)
+                            # the seek updates internal read buffers
+                            fin.seek(0, 1)
                     fin.close()
                 else:
                     out, err = popen.communicate()
             except KeyboardInterrupt:
                 self.report.keyboard_interrupt()
                 popen.wait()
-                raise KeyboardInterrupt()
+                raise
             ret = popen.wait()
         finally:
             self._popenlist.remove(popen)
@@ -196,7 +211,7 @@ class Action(object):
                 raise tox.exception.InvocationError(
                     "%s (see %s)" % (invoked, outpath), ret)
             else:
-                raise tox.exception.InvocationError("%r" % (invoked, ), ret)
+                raise tox.exception.InvocationError("%r" % (invoked,), ret)
         if not out and outpath:
             out = outpath.read()
         if hasattr(self, "commandlog"):
@@ -211,7 +226,7 @@ class Action(object):
             newargs.append(str(arg))
 
         # subprocess does not always take kindly to .py scripts
-        # so adding the interpreter here.
+        # so adding the interpreter here
         if sys.platform == "win32":
             ext = os.path.splitext(str(newargs[0]))[1].lower()
             if ext == '.py' and self.venv:
@@ -228,6 +243,14 @@ class Action(object):
                                   stdout=stdout, stderr=stderr, env=env)
 
 
+class Verbosity(object):
+    DEBUG = 2
+    INFO = 1
+    DEFAULT = 0
+    QUIET = -1
+    EXTRA_QUIET = -2
+
+
 class Reporter(object):
     actionchar = "-"
 
@@ -235,20 +258,20 @@ class Reporter(object):
         self.tw = py.io.TerminalWriter()
         self.session = session
         self._reportedlines = []
-        # self.cumulated_time = 0.0
 
     @property
     def verbosity(self):
         if self.session:
-            return self.session.config.option.verbosity
+            return (self.session.config.option.verbose_level -
+                    self.session.config.option.quiet_level)
         else:
-            return 2
+            return Verbosity.DEBUG
 
     def logpopen(self, popen, env):
         """ log information about the action.popen() created process. """
         cmd = " ".join(map(str, popen.args))
         if popen.outpath:
-            self.verbosity1("  %s$ %s >%s" % (popen.cwd, cmd, popen.outpath,))
+            self.verbosity1("  %s$ %s >%s" % (popen.cwd, cmd, popen.outpath))
         else:
             self.verbosity1("  %s$ %s " % (popen.cwd, cmd))
 
@@ -260,16 +283,16 @@ class Reporter(object):
 
     def logaction_finish(self, action):
         duration = time.time() - action._starttime
-        # self.cumulated_time += duration
         self.verbosity2("%s finish: %s after %.2f seconds" % (
             action.venvname, action.msg, duration), bold=True)
         delattr(action, '_starttime')
 
     def startsummary(self):
-        self.tw.sep("_", "summary")
+        if self.verbosity >= Verbosity.QUIET:
+            self.tw.sep("_", "summary")
 
     def info(self, msg):
-        if self.verbosity >= 2:
+        if self.verbosity >= Verbosity.DEBUG:
             self.logline(msg)
 
     def using(self, msg):
@@ -279,9 +302,6 @@ class Reporter(object):
     def keyboard_interrupt(self):
         self.error("KEYBOARDINTERRUPT")
 
-#    def venv_installproject(self, venv, pkg):
-#        self.logline("installing to %s: %s" % (venv.envconfig.envname, pkg))
-
     def keyvalue(self, name, value):
         if name.endswith(":"):
             name += " "
@@ -296,28 +316,31 @@ class Reporter(object):
         self.logline(msg, green=True)
 
     def warning(self, msg):
-        self.logline("WARNING:" + msg, red=True)
+        if self.verbosity >= Verbosity.QUIET:
+            self.logline("WARNING:" + msg, red=True)
 
     def error(self, msg):
-        self.logline("ERROR: " + msg, red=True)
+        if self.verbosity >= Verbosity.QUIET:
+            self.logline("ERROR: " + msg, red=True)
 
     def skip(self, msg):
-        self.logline("SKIPPED:" + msg, yellow=True)
+        if self.verbosity >= Verbosity.QUIET:
+            self.logline("SKIPPED:" + msg, yellow=True)
 
     def logline(self, msg, **opts):
         self._reportedlines.append(msg)
         self.tw.line("%s" % msg, **opts)
 
     def verbosity0(self, msg, **opts):
-        if self.verbosity >= 0:
+        if self.verbosity >= Verbosity.DEFAULT:
             self.logline("%s" % msg, **opts)
 
     def verbosity1(self, msg, **opts):
-        if self.verbosity >= 1:
+        if self.verbosity >= Verbosity.INFO:
             self.logline("%s" % msg, **opts)
 
     def verbosity2(self, msg, **opts):
-        if self.verbosity >= 2:
+        if self.verbosity >= Verbosity.DEBUG:
             self.logline("%s" % msg, **opts)
 
     # def log(self, msg):
@@ -335,7 +358,6 @@ class Session:
         self.report = Report(self)
         self.make_emptydir(config.logdir)
         config.logdir.ensure(dir=1)
-        # self.report.using("logdir %s" %(self.config.logdir,))
         self.report.using("tox.ini: %s" % (self.config.toxinipath,))
         self._spec2pkg = {}
         self._name2venv = {}
@@ -382,12 +404,13 @@ class Session:
 
     def runcommand(self):
         self.report.using("tox-%s from %s" % (tox.__version__, tox.__file__))
+        verbosity = (self.report.verbosity > Verbosity.DEFAULT)
         if self.config.option.showconfig:
             self.showconfig()
         elif self.config.option.listenvs:
-            self.showenvs(all_envs=False, description=self.config.option.verbosity > 0)
+            self.showenvs(all_envs=False, description=verbosity)
         elif self.config.option.listenvs_all:
-            self.showenvs(all_envs=True, description=self.config.option.verbosity > 0)
+            self.showenvs(all_envs=True, description=verbosity)
         else:
             return self.subcommand_test()
 
@@ -404,7 +427,17 @@ class Session:
     def _makesdist(self):
         setup = self.config.setupdir.join("setup.py")
         if not setup.check():
-            raise tox.exception.MissingFile(setup)
+            self.report.error(
+                "No setup.py file found. The expected location is:\n"
+                "  %s\n"
+                "You can\n"
+                "  1. Create one:\n"
+                "     https://packaging.python.org/tutorials/distributing-packages/#setup-py\n"
+                "  2. Configure tox to avoid running sdist:\n"
+                "     http://tox.readthedocs.io/en/latest/example/general.html"
+                "#avoiding-expensive-sdist" % setup
+            )
+            raise SystemExit(1)
         action = self.newaction(None, "packaging")
         with action:
             action.setactivity("sdist-make", setup)
@@ -442,9 +475,9 @@ class Session:
     def setupenv(self, venv):
         if venv.envconfig.missing_subs:
             venv.status = (
-                "unresolvable substitution(s): %s. "
-                "Environment variables are missing or defined recursively." %
-                (','.join(["'%s'" % m for m in venv.envconfig.missing_subs])))
+                    "unresolvable substitution(s): %s. "
+                    "Environment variables are missing or defined recursively." %
+                    (','.join(["'%s'" % m for m in venv.envconfig.missing_subs])))
             return
         if not venv.matching_platform():
             venv.status = "platform mismatch"
@@ -452,6 +485,7 @@ class Session:
         action = self.newaction(venv, "getenv", venv.envconfig.envdir)
         with action:
             venv.status = 0
+            default_ret_code = 1
             envlog = self.resultlog.get_envlog(venv.name)
             try:
                 status = venv.update(action=action)
@@ -459,18 +493,25 @@ class Session:
                 if e.args[0] != 2:
                     raise
                 status = (
-                    "Error creating virtualenv. Note that spaces in paths are "
-                    "not supported by virtualenv. Error details: %r" % e)
+                        "Error creating virtualenv. Note that spaces in paths are "
+                        "not supported by virtualenv. Error details: %r" % e)
             except tox.exception.InvocationError as e:
                 status = (
-                    "Error creating virtualenv. Note that some special "
-                    "characters (e.g. ':' and unicode symbols) in paths are "
-                    "not supported by virtualenv. Error details: %r" % e)
+                        "Error creating virtualenv. Note that some special "
+                        "characters (e.g. ':' and unicode symbols) in paths are "
+                        "not supported by virtualenv. Error details: %r" % e)
+            except tox.exception.InterpreterNotFound as e:
+                status = e
+                if self.config.option.skip_missing_interpreters:
+                    default_ret_code = 0
             if status:
                 commandlog = envlog.get_commandlog("setup")
-                commandlog.add_command(["setup virtualenv"], str(status), 1)
+                commandlog.add_command(["setup virtualenv"], str(status), default_ret_code)
                 venv.status = status
-                self.report.error(str(status))
+                if default_ret_code == 0:
+                    self.report.skip(str(status))
+                else:
+                    self.report.error(str(status))
                 return False
             commandpath = venv.getcommandpath("python")
             envlog.set_python_info(commandpath)
@@ -566,24 +607,23 @@ class Session:
                     else:
                         self.installpkg(venv, path)
 
-                # write out version dependency information
-                action = self.newaction(venv, "envreport")
-                with action:
-                    args = venv.envconfig.list_dependencies_command
-                    output = venv._pcall(args,
-                                         cwd=self.config.toxinidir,
-                                         action=action)
-                    # the output contains a mime-header, skip it
-                    output = output.split("\n\n")[-1]
-                    packages = output.strip().split("\n")
-                    action.setactivity("installed", ",".join(packages))
-                    envlog = self.resultlog.get_envlog(venv.name)
-                    envlog.set_installed(packages)
-
+                self.runenvreport(venv)
                 self.runtestenv(venv)
         retcode = self._summary()
         return retcode
 
+    def runenvreport(self, venv):
+        """
+        Run an environment report to show which package
+        versions are installed in the venv
+        """
+        action = self.newaction(venv, "envreport")
+        with action:
+            packages = self.hook.tox_runenvreport(venv=venv, action=action)
+        action.setactivity("installed", ",".join(packages))
+        envlog = self.resultlog.get_envlog(venv.name)
+        envlog.set_installed(packages)
+
     def runtestenv(self, venv, redirect=False):
         if not self.config.option.notest:
             if venv.status:
@@ -649,7 +689,7 @@ class Session:
     def showenvs(self, all_envs=False, description=False):
         env_conf = self.config.envconfigs  # this contains all environments
         default = self.config.envlist  # this only the defaults
-        extra = sorted([e for e in env_conf if e not in default]) if all_envs else []
+        extra = sorted(e for e in env_conf if e not in default) if all_envs else []
         if description:
             self.report.line('default environments:')
             max_length = max(len(env) for env in (default + extra))
@@ -657,10 +697,11 @@ class Session:
         def report_env(e):
             if description:
                 text = env_conf[e].description or '[no description]'
-                msg = '{0} -> {1}'.format(e.ljust(max_length), text).strip()
+                msg = '{} -> {}'.format(e.ljust(max_length), text).strip()
             else:
                 msg = e
             self.report.line(msg)
+
         for e in default:
             report_env(e)
         if all_envs and extra:
@@ -677,7 +718,7 @@ class Session:
             stdout=subprocess.PIPE,
         )
         out, _ = proc.communicate()
-        versions.append('virtualenv-{0}'.format(out.decode('UTF-8').strip()))
+        versions.append('virtualenv-{}'.format(out.decode('UTF-8').strip()))
         self.report.keyvalue("tool-versions:", " ".join(versions))
 
     def _resolve_pkg(self, pkgspec):
